import asyncio
import functools
import sys

import pytest

from aiortsp.rtsp.auth import DigestAuth
from aiortsp.rtsp.connection import RTSPConnection
from aiortsp.rtsp.errors import RTSPResponseError
from aiortsp.rtsp.parser import RTSPParser


async def handle_client(queue, client_reader, client_writer):

    count = 0
    parser = RTSPParser()

    while True:
        data = await client_reader.read(10000)

        for msg in parser.parse(data):
            if count == 0:
                assert msg.content == 'Dumb ass!'
                assert msg.headers['content-length'] == str(len(b'Dumb ass!'))

            resp = f"""RTSP/1.0 200 OK\r
CSeq: {msg.cseq}\r
\r
"""
            if count == 0:
                resp += f"""ANNOUNCE rtsp://foo/bar.avi RTSP/1.0\r
CSeq: {msg.cseq + 1}\r
\r
"""
            client_writer.write(resp.encode())
            count += 1
            queue.put_nowait(msg)


@pytest.mark.asyncio
async def test_client_connection():
    queue = asyncio.Queue()

    server = await asyncio.start_server(functools.partial(handle_client, queue), '127.0.0.1', 5554)
    try:
        async with RTSPConnection('127.0.0.1', 5554, 'foo', 'bar') as conn:

            # Send an OPTIONS request
            resp = await conn.send_request('OPTIONS', '*', timeout=2, body=b'Dumb ass!')
            assert resp
            assert resp.status == 200

            # Ensure the other side have seen a request
            req = await asyncio.wait_for(queue.get(), 2)
            assert req.type == 'request'
            assert req.method == 'OPTIONS'
            assert req.cseq == resp.cseq

            # Then it should have sent an announce, and got a 551 reply
            req = await asyncio.wait_for(queue.get(), 2)
            assert req.type == 'response'
            assert req.status == 551
            assert req.cseq == resp.cseq + 1
    finally:
        server.close()


async def handle_client_auth(client_reader, client_writer):
    parser = RTSPParser()

    while True:
        data = await client_reader.read(10000)
        for msg in parser.parse(data):
            authorized = False
            if 'authorization' in msg.headers:
                # Check it!
                params = DigestAuth._parse_digest_header(msg.headers['authorization'].split(' ', 1)[-1])
                print("PARAMS", params)
                if params.get('nonce') == '0024e47aY398109708de9ccd8056c58a068a59540a99d3' and \
                    params.get('realm') == 'AXIS_ACCC8E000AA9' and \
                    params.get('opaque') == '123soleil' and \
                    params.get('uri') == 'rtsp://cam/axis-media/media.amp' and \
                    params.get('username') == 'root' and \
                    params.get('response') == '7daaf0f4e40fdff42cff28260f37914d':
                    authorized = True

            if authorized:
                client_writer.write("""RTSP/1.0 200 OK\r\nCSeq: {}\r\n\r\n""".format(msg.headers.get('cseq', 0)).encode())
            else:
                client_writer.write("""RTSP/1.0 401 Unauthorized\r
CSeq: {}\r
WWW-Authenticate: Digest realm="AXIS_ACCC8E000AA9", nonce="0024e47aY398109708de9ccd8056c58a068a59540a99d3", opaque="123soleil"\r
\r
""".format(msg.headers.get('cseq', 0)).encode())


@pytest.mark.asyncio
async def test_client_auth_no_credentials():
    server = await asyncio.start_server(handle_client_auth, '127.0.0.1', 5554)
    try:
        async with RTSPConnection('127.0.0.1', 5554) as conn:
            with pytest.raises(RTSPResponseError):
                await conn.send_request('DESCRIBE', 'rtsp://cam/axis-media/media.amp', timeout=2)
    finally:
        server.close()


@pytest.mark.asyncio
async def test_client_auth_invalid_credentials():
    server = await asyncio.start_server(handle_client_auth, '127.0.0.1', 5554)
    try:
        async with RTSPConnection('127.0.0.1', 5554, username='toto', password='hello') as conn:
            with pytest.raises(RTSPResponseError):
                await conn.send_request('DESCRIBE', 'rtsp://cam/axis-media/media.amp', timeout=2)
    finally:
        server.close()


@pytest.mark.asyncio
async def test_client_auth():
    server = await asyncio.start_server(handle_client_auth, '127.0.0.1', 5554)
    try:
        async with RTSPConnection('127.0.0.1', 5554, username='root', password='admin123') as conn:
            resp = await conn.send_request('DESCRIBE', 'rtsp://cam/axis-media/media.amp', timeout=2)
            assert resp
            assert resp.status == 200
    finally:
        server.close()



import asyncio

import pytest

from aiortsp.rtsp.connection import RTSPConnection
from aiortsp.rtsp.parser import RTSPParser, RTSPRequest
from aiortsp.rtsp.session import RTSPMediaSession
from aiortsp.transport import TCPTransport


async def handle_client_auth(client_reader, client_writer):
    parser = RTSPParser()
    playing = False

    while True:
        data = await client_reader.read(10000)

        if not data:
            break

        for msg in parser.parse(data):
            assert isinstance(msg, RTSPRequest)
            print('MSG', msg)
            response = 'RTSP/1.0 200 OK\r\nCSeq: {}\r\n'.format(msg.cseq)
            if msg.method == 'OPTIONS':
                response += 'Public: DESCRIBE, SETUP, TEARDOWN, PLAY, PAUSE\r\n'
            elif msg.method == 'DESCRIBE':
                response += 'Content-Type: application/sdp\r\n'
                response += 'Content-Base: rtsp://cam/media.sdp/\r\n'
                response += 'Server: Dummy Test server\r\n'
                response += 'Content-Length: 617\r\n'
                response += '\r\n'
                response += 'v=0\r\n'
                response += 'o=- 17428449743163035608 1 IN IP4 10.10.0.77\r\n'
                response += 's=Session streamed with GStreamer\r\n'
                response += 'i=rtsp-server\r\n'
                response += 't=0 0\r\n'
                response += 'a=tool:GStreamer\r\n'
                response += 'a=type:broadcast\r\n'
                response += 'a=range:npt=now-\r\n'
                response += 'a=control:rtsp://10.10.0.77/axis-media/media.amp\r\n'
                response += 'm=video 0 RTP/AVP 96\r\n'
                response += 'c=IN IP4 0.0.0.0\r\n'
                response += 'b=AS:50000\r\n'
                response += 'a=rtpmap:96 H264/90000\r\n'
                response += 'a=fmtp:96 packetization-mode=1;profile-level-id=4d0029;sprop-parameter-sets=Z00AKeKQCgC3YC3AQEBpB4kRUA==,aO48gA==\r\n'
                response += 'a=ts-refclk:local\r\n'
                response += 'a=mediaclk:sender\r\n'
                response += 'a=control:rtsp://10.10.0.77/axis-media/media.amp/stream=0\r\n'
                response += 'a=framerate:25.000000\r\n'
                response += 'a=transform:1.000000,0.000000,0.000000;0.000000,1.000000,0.000000;0.000000,0.000000,1.000000\r\n'
            elif msg.method == 'SETUP':
                response += 'Transport: RTP/AVP/TCP;unicast;interleaved=0-1;ssrc=E6EC9FEF;mode="PLAY"\r\n'
                response += 'Session: 2sY7Pd2EPx8JY50-;timeout=60\r\n'
            elif msg.method == 'PLAY':
                playing = True
            response += '\r\n'
            print('RESPONSE', response)
            client_writer.write(response.encode())

            if playing:
                # Send 2 RTP packets
                rtp = bytearray.fromhex('2400002080605eaac639ab5e13cd9b86674d0029e29019077f1180b7010101a41e244540'
                                        '2400001080605eabc639ab5e13cd9b8668ee3c80')
                client_writer.write(rtp)

                # Send an SR
                sr = bytearray.fromhex('80c8000677ae8d65e051bc2bea33b0001fa8034c0000000000000000')
                msg = bytearray([ord('$'), 1, 0, len(sr)])
                msg.extend(sr)
                client_writer.write(msg)


@pytest.mark.asyncio
async def test_session():
    server = await asyncio.start_server(handle_client_auth, '127.0.0.1', 5554)
    try:
        async with RTSPConnection('127.0.0.1', 5554, timeout=2) as conn:
            async with TCPTransport(conn) as transport:
                async with RTSPMediaSession(conn, 'rtsp://cam/media.sdp', transport) as sess:
                    rtcp = sess.stats.build_rtcp()
                    assert rtcp is None
                    await sess.play()
                    await asyncio.sleep(0.2)
                    rtcp = sess.stats.build_rtcp()
                    assert rtcp
                    assert sess.stats.received == 2
    finally:
        server.close()



import asyncio
import logging
import sys

import pytest
from dpkt.rtp import RTP

from aiortsp.rtsp.reader import RTSPReader
from tests.test_session import handle_client_auth


@pytest.mark.asyncio
async def test_reader():
    count = 0
    server = await asyncio.start_server(handle_client_auth, '127.0.0.1', 5554)
    try:
        async with RTSPReader('rtspt://127.0.0.1:5554/media.sdp', timeout=2) as reader:
            async for pkt in reader.iter_packets():
                assert isinstance(pkt, RTP)
                count += 1

                if count >= 2:
                    server.close()

        assert count == 2
    finally:
        server.close()


@pytest.mark.asyncio
async def test_reader_reconnect():
    logging.basicConfig()
    count = 0

    server = await asyncio.start_server(handle_client_auth, '127.0.0.1', 5554)
    try:
        async with RTSPReader('rtspt://127.0.0.1:5554/media.sdp', run_loop=True, timeout=2, log_level=10) as reader:
            async for pkt in reader.iter_packets():
                print('PKT', len(pkt))
                assert isinstance(pkt, RTP)
                count += 1

                if count == 2:
                    # closing connection to test reconnect
                    reader.connection.close()

                if count == 4:
                    break
    finally:
        server.close()



"""
RTSPS example using the RTSPReader
"""

import asyncio
import contextlib
import logging
import ssl
import sys

from aiortsp.rtsp.reader import RTSPReader

logging.getLogger().setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
logging.getLogger().addHandler(handler)


async def main():
    import argparse
    from urllib.parse import urlparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-l", "--logging", type=int, default=20, help="RTSP url")
    parser.add_argument("url", help="RTSPS url")
    args = parser.parse_args()

    p_url = urlparse(args.url)

    if p_url.scheme != "rtsps":
        logging.error("this example only supports rtsp:// URLs")
        sys.exit(1)

    ssl_ctx = ssl.create_default_context()
    ssl_ctx.check_hostname = False
    ssl_ctx.verify_mode = ssl.CERT_NONE
    # Open a reader (which means RTSP connection, then media session)
    async with RTSPReader(args.url, log_level=args.logging, run_loop=True, ssl=ssl_ctx) as reader:
        # Iterate on RTP packets
        async for pkt in reader.iter_packets():
            logging.info("PKT %s", (pkt.cc, pkt.m, pkt.pt, pkt.seq, pkt.ts, pkt.ssrc, len(pkt.data)))
    logging.info("... all done...")


if __name__ == "__main__":
    with contextlib.suppress(KeyboardInterrupt):
        asyncio.run(main())



"""
TCP Interleaved transport.
"""
import logging

from aiortsp.rtcp.parser import RTCP
from aiortsp.rtsp.errors import RTSPError
from aiortsp.rtsp.parser import RTSPBinary
from .base import RTPTransport

_logger = logging.getLogger('rtp.session')

DEFAULT_BUFFER_SIZE = 4 * 1024 * 1024


class TCPTransport(RTPTransport):
    """
    TCP Transport.
    --------------

    Uses connection directly
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.rtp_idx = None
        self.rtcp_idx = None

        self.receive_buffer = kwargs.get('receive_buffer', DEFAULT_BUFFER_SIZE)
        self.send_buffer = kwargs.get('send_buffer', DEFAULT_BUFFER_SIZE)

    async def prepare(self):
        self.rtp_idx = self.connection.register_binary_handler(self.handle_rtp_bin)
        self.rtcp_idx = self.connection.register_binary_handler(self.handle_rtcp_bin)
        self.logger.info(
            'receiving interleaved RTP (%s) and RTCP (%s)',
            self.rtp_idx, self.rtcp_idx
        )

    @property
    def running(self) -> bool:
        """
        True if both RTP and RTCP sinks are running.
        """
        return self.connection.running

    def close(self, error=None):
        """
        Perform cleanup, which by default is closing both sinks.
        """
        super().close(error)
        self.connection.close()

    def handle_rtcp_bin(self, binary: RTSPBinary):
        """
        Handle interleaved data registered as RTCP
        """
        self.handle_rtcp_data(binary.data)

    def handle_rtp_bin(self, binary: RTSPBinary):
        """
        Handle interleaved data registered as RTP
        """
        self.handle_rtp_data(binary.data)

    def on_transport_request(self, headers: dict):
        headers['Transport'] = f'RTP/AVP/TCP;unicast;interleaved={self.rtp_idx}-{self.rtcp_idx}'

    def on_transport_response(self, headers: dict):
        if 'transport' not in headers:
            raise RTSPError('error on SETUP: Transport not found')

        fields = self.parse_transport_fields(headers['transport'])

        assert fields.get('interleaved') == f'{self.rtp_idx}-{self.rtcp_idx}', 'invalid returned interleaved header'

    async def send_rtcp_report(self, rtcp: RTCP):
        self.connection.send_binary(self.rtcp_idx, bytes(rtcp))



import pytest

from aiortsp.rtsp.auth import DigestAuth, BasicAuth
from aiortsp.rtsp.parser import RTSPResponse


@pytest.mark.parametrize('user, passwd, method, url, req, repl', [

    ('root', 'admin123', 'DESCRIBE', 'rtsp://cam/axis-media/media.amp', {
        # Challenge
        'realm': "AXIS_ACCC8E000AA9",
        'nonce': "0024e47aY398109708de9ccd8056c58a068a59540a99d3"
    }, {
        # Response
        "username": "root",
        'realm': "AXIS_ACCC8E000AA9",
        'nonce': "0024e47aY398109708de9ccd8056c58a068a59540a99d3",
        'uri': "rtsp://cam/axis-media/media.amp",
        'response': "7daaf0f4e40fdff42cff28260f37914d",
    }),

    ('test', 'test123', 'DESCRIBE', 'rtsp://recorder:654/00000001-0000-babe-0000-accc8e000aa7/live', {
        # Challenge
        'realm': "media@genetec.com",
        'nonce': "900fa9ee25fb4d5e919fa17c2cd032f7",
        'opque': "0bd45fa6a89e4873a4c80ecf6287611f",
        'qop': "auth",
        'stale': "FALSE",
        'algorithm': "MD5"
    }, {
        # Response
        "username": "test",
        'realm': "media@genetec.com",
        'nonce': "900fa9ee25fb4d5e919fa17c2cd032f7",
        'uri': "rtsp://recorder:654/00000001-0000-babe-0000-accc8e000aa7/live",
        'response': "85ab8f66f2930845b1ed05742a2ad8b4",
    }),
])
def test_digest(user, passwd, method, url, req, repl):
    auth = DigestAuth(username=user, password=passwd)

    resp = RTSPResponse()

    auth_header = ', '.join(f'{k}="{v}"' for k, v in req.items())
    data, done = resp.feed(
        """RTSP/1.0 401 Unauthorized\r\nCSeq: 0\r\nWWW-Authenticate: Digest {}\r\n\r\n""".format(auth_header).encode())

    assert not data
    assert done

    assert resp.status == 401

    retry = auth.handle_401(resp.headers)
    assert retry is True

    # Force the case where auth is to be redone
    auth.handle_ok(resp.headers)
    retry = auth.handle_401(resp.headers)
    assert retry is True

    auth_args = auth._prepare_digest_header(method=method, url=url)

    assert auth_args == repl

    # Force the case where auth is to be redone
    auth.handle_ok({
        'authentication-info': 'qop="auth", nextnonce="deadb00b"'
    })
    assert auth.info['nonce'] == 'deadb00b'


@pytest.mark.parametrize('user, passwd, method, url, req, repl', [

    # Against Axis
    ('root', 'admin123', 'DESCRIBE', 'rtsp://cam/axis-media/media.amp', {
        # Challenge
        'realm': "AXIS_ACCC8E000AA9",
        'nonce': "0024e47aY398109708de9ccd8056c58a068a59540a99d3"
    }, 'Basic cm9vdDphZG1pbjEyMw=='),
])
def test_basic(user, passwd, method, url, req, repl):
    auth = BasicAuth(username=user, password=passwd)

    resp = RTSPResponse()

    auth_header = ', '.join(f'{k}="{v}"' for k, v in req.items())
    data, done = resp.feed(
        """RTSP/1.0 401 Unauthorized\r\nCSeq: 0\r\nWWW-Authenticate: Basic {}\r\n\r\n""".format(auth_header).encode())

    assert not data
    assert done

    assert resp.status == 401

    retry = auth.handle_401(resp.headers)
    assert retry is True

    # Force the case where auth is to be redone
    auth.handle_ok(resp.headers)
    retry = auth.handle_401(resp.headers)
    assert retry is True

    headers = {}
    auth.make_auth(method, url, headers)

    assert 'Authorization' in headers
    assert headers['Authorization'] == repl



"""
RTP Transport base class.
---------------------
"""
import asyncio
import logging
import traceback
from time import time
from typing import Optional, Set

from dpkt.rtp import RTP

from aiortsp.rtcp.parser import RTCP
from aiortsp.rtcp.stats import RTCPStats

_logger = logging.getLogger('rtp.session')

DEFAULT_BUFFER_SIZE = 4 * 1024 * 1024


class RTPTransportClient:
    """
    RTP Transport Client class.
    - To be subscribed to the transport
    - Implementing wanted callbacks
    """

    def handle_rtp(self, rtp: RTP):
        """Handle an incoming RTP packet"""

    def handle_rtcp(self, rtcp: RTCP):
        """Handle an incoming RTP packet"""

    def handle_closed(self, error: Optional[Exception]):
        """Handle closing"""


class RTPTransport:
    """
    Base (abstract) RTP Transport class
    """

    def __init__(self, connection, *, logger=None, timeout=10):
        self.connection = connection
        self.clients: Set[RTPTransportClient] = set()
        self.stats = RTCPStats()
        self.logger = logger or _logger
        self.result = asyncio.Future()
        # Just to be sure we retrieve the exception at least once if client doesn't
        self.result.add_done_callback(lambda fut: fut.exception())
        self.timeout = timeout
        self.paused = False
        self._rtcp_loop = self._timeout_loop = None

    def subscribe(self, client: RTPTransportClient):
        """
        Add a subscription to RTP frames
        """
        self.clients.add(client)

    def unsubscribe(self, client: RTPTransportClient):
        """
        Remove a subscription to RTP frames
        """
        self.clients.remove(client)

    def handle_rtcp_data(self, data: bytes):
        """
        An RTCP report was received.
        """
        rtcp = RTCP.unpack(data)
        self.logger.debug('received RTCP report: %s', rtcp)

        self.stats.handle_rtcp(rtcp)

        if self._rtcp_loop is None:
            # This is the first packet ever!
            self._rtcp_loop = asyncio.ensure_future(self.rtcp_loop())

        for client in self.clients:
            try:
                client.handle_rtcp(rtcp)
            except Exception as ex:  # pylint: disable=broad-except
                self.logger.error('error on RTCP client callback: %r', ex)

    def handle_rtp_data(self, data: bytes):
        """
        An RTP packet was received.
        """
        rtp = RTP(data)

        self.stats.update(rtp)

        for client in self.clients:
            try:
                client.handle_rtp(rtp)
            except Exception as ex:  # pylint: disable=broad-except
                self.logger.error('error on RTP client callback: %r', ex)

    @staticmethod
    def parse_transport_fields(header) -> dict:
        """
        Parse Transport request/response
        """
        res = {}
        fields_ = header.split(';')
        for field in fields_:
            if '=' in field:
                k, v = field.split('=', 1)
            else:
                k, v = field, None

            res[k.strip()] = v.strip() if v is not None else None

        return res

    async def rtcp_loop(self):
        """
        Handling report sending and timeout
        """
        initial = True
        while self.running:
            delay = self.stats.rtcp_interval(initial)
            initial = False
            self.logger.debug('sleeping for RTCP delay: %s', delay)
            await asyncio.sleep(delay)

            try:
                rtcp = self.stats.build_rtcp()

                if not rtcp:
                    self.logger.warning('no RTCP report available yet (are we receiving anything?)')
                    continue
            except Exception as ex:  # pylint: disable=broad-except
                self.logger.error('unable to build RTCP report: %r', ex)
                traceback.print_exc()
                continue

            try:
                self.logger.debug('sending RTCP report: %s', rtcp)
                await self.send_rtcp_report(rtcp)
            except asyncio.CancelledError:
                break
            except Exception as ex:  # pylint: disable=broad-except
                self.logger.error('unable to send RTCP report: %r', ex)

    async def timeout_loop(self):
        """
        Handling report sending and timeout
        """
        sleep_duration = self.timeout
        while self.running:
            await asyncio.sleep(sleep_duration)

            # Check if (and when) we received anything
            diff = time() - self.stats.last_received
            if not self.paused and diff > self.timeout:
                self.logger.error('no RTP received for %s seconds: closing', self.timeout)
                return self.close(TimeoutError('no data'))

            sleep_duration = max(self.timeout - diff, 1)

    def close(self, error=None):
        """
        Called when transport needs closing
        """
        if not self.result.done():
            if error:
                self.result.set_exception(error)
            else:
                self.result.set_result(False)

        if self._rtcp_loop:
            self._rtcp_loop.cancel()
            self._rtcp_loop = None

        if self._timeout_loop and not self._timeout_loop.done():
            self._timeout_loop.cancel()
            self._timeout_loop = None

        for client in self.clients:
            try:
                client.handle_closed(error)
            except Exception as ex:  # pylint: disable=broad-except
                self.logger.error('error on RTP client callback: %r', ex)

    async def __aenter__(self):
        await self.prepare()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.cleanup()
        self.close(exc_val)

    def __enter__(self):
        raise RuntimeError('did you mean `async with`?')

    def __exit__(self, exc_type, exc_val, exc_tb):
        raise RuntimeError('should never be called')

    def on_transport_request(self, headers: dict):
        """
        Given a setup request headers dict,
        add whatever headers are necessary, usually the 'Transport' header.
        """
        raise NotImplementedError

    def on_transport_response(self, headers: dict):
        """
        Given a setup response headers dict,
        Allow the transport to read whatever is necessary.
        """
        raise NotImplementedError

    async def prepare(self):
        """
        Called before setting up the media session.
        Any
        :return:
        """

    async def cleanup(self):
        """
        Called before setting up the media session.
        Any
        :return:
        """

    async def warmup(self):
        """
        Called before playing.
        Whatever should be done before playing should be done here.
        """
        if self.timeout:
            self._timeout_loop = asyncio.ensure_future(self.timeout_loop())

    @property
    def running(self) -> bool:
        """
        Tells if the transport is still open and running
        """
        raise NotImplementedError

    async def send_rtcp_report(self, rtcp: RTCP):
        """
        Send an RTCP report back to the server
        """
        raise NotImplementedError



"""
Asyncio RTSP Connection module
This is the main interesting part for the user.
"""

import asyncio
import logging
import traceback
from typing import Callable

from aiortsp.__version__ import __version__
from .auth import BasicAuth, DigestAuth
from .errors import RTSPResponseError, RTSPConnectionError, RTSPTimeoutError
from .parser import RTSPParser, RTSPRequest, RTSPResponse, RTSPBinary

_logger = logging.getLogger('rtsp_client')


LINE_SPLIT_STR = '\r\n'
HEADER_END_STR = LINE_SPLIT_STR * 2
USER_AGENT = f'aiortsp/{__version__}'


class RTSPConnection(asyncio.Protocol):
    """
    Creates an RTSP connection for asyncio usage.

    It's as easy as:

    async with RTSPConnection(...) as conn:
          # Here you go, do your RTSP stuff
          resp = await conn.send_request('DESCRIBE', url)
    # Cleans properly the connection before leaving the context

    """

    def __init__(self, host, port, username=None, password=None, accept_auth=None, logger=None, ssl=None, timeout=10):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.accept_auth = [auth.lower() for auth in accept_auth] if accept_auth else ['basic', 'digest']
        self.default_timeout = timeout
        self.logger = logger or _logger
        self.ssl = ssl

        self._transport = None
        self.result = asyncio.Future()
        self.pending_msg = None
        self.active_requests = {}
        self._cseq = 1
        self._auth = None
        self.parser = RTSPParser()
        self.binary_handlers = {}

    async def __aenter__(self):
        await self.prepare()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.close()

    async def prepare(self):
        """
        Prepare connection
        :return:
        """
        loop = asyncio.get_event_loop()
        try:
            await asyncio.wait_for(
                loop.create_connection(lambda: self, self.host, self.port, ssl = self.ssl),
                self.default_timeout
            )
        except (asyncio.TimeoutError, OSError) as to:
            raise RTSPConnectionError(f'Unable to connect to {self.host}:{self.port}') from to

    def close(self):
        """
        Close connection
        """
        if self._transport:
            self._transport.close()

    def register_binary_handler(self, callback: Callable) -> int:
        """
        Register a binary callback. Return the ID which will be used for given protocol
        """
        idx = next(i for i in range(257) if i not in self.binary_handlers)

        assert idx < 256, 'not any binary handle left'

        self.binary_handlers[idx] = callback

        return idx

    def connection_made(self, transport):
        """Conforms asyncio.Protocol"""
        self._transport = transport

    def connection_lost(self, exc):
        """Conforms asyncio.Protocol"""
        self.logger.info('connection to RTSP server %s:%s closed (error: %s)', self.host, self.port, exc)
        self._transport = None
        if not self.result.done():
            if exc:
                error = RTSPConnectionError(f'RTSP connection lost: {exc}')
                error.__cause__ = exc
                self.result.set_exception(error)
            else:
                self.result.set_result('ok')

        # Close any pending request
        error = self.result.exception() if not self.result.cancelled() else None
        for request in self.active_requests.values():  # type: asyncio.Future
            if not request.done():
                request.set_exception(error or RTSPConnectionError('connection closed'))

    @property
    def running(self) -> bool:
        """
        Tells if currently running, ie if we have an open transport.
        """
        return self._transport is not None

    def on_binary(self, binary: RTSPBinary):
        """Handler for binary data received"""
        if binary.id in self.binary_handlers:
            # Call handler
            self.binary_handlers[binary.id](binary)
        else:
            self.logger.debug('BINARY data (%s bytes): %s', binary.length, binary.content)

    def on_response(self, response: RTSPResponse):
        """Handler for response received"""
        self.logger.debug('RESPONSE received: %s\n%s', response, response.content)

    def on_request(self, request: RTSPRequest):
        """
        Handle a request from server. Override for more fancy controls
        """
        self.logger.warning('request message received during session:\n%s', request.content)
        # @TODO We do not support requests for now: 551
        self.send_response(request, 551, 'Option not supported')

    def data_received(self, data: bytes):
        """Parse and distribute received messages"""
        try:
            # self.logger.debug('<<< receiving data: %s', data)

            for msg in self.parser.parse(data):
                # self.logger.debug('message done: %s', msg)
                if msg.type == 'response':
                    self.on_response(msg)

                    if msg.cseq in self.active_requests:
                        self.active_requests[msg.cseq].set_result(msg)

                elif msg.type == 'binary':
                    self.on_binary(msg)

                elif msg.type == 'request':
                    self.on_request(msg)

            return

        except Exception as ex:  # pylint: disable=broad-except
            self.logger.error('error on received data: %s\n%s', ex.__class__.__name__, data)

            error = RTSPConnectionError('invalid data received from RTSP connection')
            error.__cause__ = ex
            self.result.set_exception(error)

            if self._transport:
                self._transport.close()
            traceback.print_exc()

    def _next_seq(self):
        cseq = self._cseq
        self._cseq += 1
        return cseq

    def handle_401(self, resp: RTSPResponse):
        """
        Handle a 401 (Unauthorized) message.
        :param resp: Response from server containing challenge (digest, basic)
        :return: True if client is authorized to retry
        """
        if not self._auth and self.username and self.password:
            # # No authentication selected yet: use one!
            # if not (self.username and self.password):
            #     raise RTSPResponseError('No valid Authentication provided (username/password)', resp)

            # Check what is supported
            www_auth = resp.headers.get('www-authenticate')

            if not www_auth:
                raise RTSPResponseError('Invalid 401 response received (no www-authenticate)', resp)

            if not isinstance(www_auth, list):
                www_auth = [www_auth]

            self.logger.debug('authorization attempt, allowed: %s, proposed: %s', self.accept_auth, www_auth)

            if any(a.startswith('Basic ') for a in www_auth) and 'basic' in self.accept_auth:
                self.logger.debug('selecting BASIC authentication')
                self._auth = BasicAuth(self.username, self.password)
            elif any(a.startswith('Digest ') for a in www_auth) and 'digest' in self.accept_auth:
                self.logger.debug('selecting DIGEST authentication')
                self._auth = DigestAuth(self.username, self.password)

        if self._auth:
            return self._auth.handle_401(resp.headers)

        return False

    def handle_ok(self, resp: RTSPResponse):
        """
        Handle an OK (in terms of authentication) response.
        There may be extra pieces of information for Auth.
        :param resp:
        :return:
        """
        if self._auth:
            self._auth.handle_ok(resp.headers)

    def send_message(self, msg, cseq, headers, body: bytes = None):
        """
        Send a 'message' (request or reply) with given cseq and headers
        """
        if not self._transport:
            self.logger.error('transport is closed')
            return

        # Always write CSeq first
        msg += f'{LINE_SPLIT_STR}CSeq: {cseq}'

        headers['User-Agent'] = USER_AGENT

        if body:
            headers['Content-Length'] = len(body)

        for k, v in headers.items():
            msg += f'{LINE_SPLIT_STR}{k}: {v}'

        msg += HEADER_END_STR  # End of headers

        data = msg.encode()

        if body:
            data += body

        self._transport.write(data)
        self.logger.debug('>>> sending msg:\n%s\n', msg)

    async def send_request(self, method, url, headers=None, timeout=None, body: bytes = None) -> RTSPResponse:
        """
        Send an RTSP request.
        :param method: RTSP method to be sent
        :param url: URL to be given in the RTSP request header
        :param headers: dict of optional headers to add
        :param timeout: timeout for getting a response
        :param body: Content body. If specified, a 'Content-Type' header should be added.
        :return: RTSPResponse
        """
        request = f'{method} {url} RTSP/1.0'

        if headers is None:
            headers = {}

        if self._auth:
            self._auth.make_auth(method, url, headers)

        cseq = self._next_seq()
        self.active_requests[cseq] = asyncio.Future()

        try:
            self.send_message(request, cseq, headers, body)

            resp = await asyncio.wait_for(self.active_requests[cseq], timeout or self.default_timeout)

            if resp.status == 401:
                self.logger.debug('unauthorized; see if we can try to authenticate')
                retry = self.handle_401(resp)
                if retry:
                    return await self.send_request(method, url, headers, timeout, body)
            else:
                # Response was successful (or at least not unauthorized...)
                self.handle_ok(resp)

            if not 200 <= resp.status < 300:
                raise RTSPResponseError(f'RTSP request error for {method} {url}', resp)

            return resp

        except asyncio.TimeoutError as to:
            raise RTSPTimeoutError('RTSP server failed to answer in time') from to

        finally:
            # Always cleanup the active request we had registered
            self.active_requests.pop(cseq, None)

    def send_response(self, request: RTSPRequest, code, msg, headers=None):
        """
        Send a response message to given request
        """
        if not self._transport:
            self.logger.error('transport is closed')
            return

        response = f'RTSP/1.0 {code} {msg}'

        if headers is None:
            headers = {}

        if 'session' in request.headers:
            headers['Session'] = request.headers['session']

        self.send_message(response, request.cseq, headers)

    def send_binary(self, idx: int, data: bytes):
        """
        Send a binary interleaved packet
        """
        if not self._transport:
            self.logger.error('transport is closed')
            return

        assert 0 <= idx < 256, f'invalid binary index: {idx}'

        m_len = len(data)

        msg = bytearray([ord('$'), idx, (m_len & 0xFF00) >> 8, m_len & 0xFF])
        msg += data

        self._transport.write(msg)
        self.logger.debug('>>> sending binary: %s', msg)



"""
RTSP Authentication module.
---------------------------

Implements Basic and Digest authentication.
"""

import hashlib

from base64 import b64encode
from typing import Callable
from urllib.request import parse_http_list


DIGEST_METHODS = {
    'MD5': hashlib.md5,
    'SHA': hashlib.sha1,
    'SHA-256': hashlib.sha256,
    'SHA-512': hashlib.sha512
}


class Auth:
    """
    Base class for authentication
    """

    def __init__(self, max_retry=1):
        self.max_retry = max_retry
        self.retry_count = 0

    def handle_ok(self, headers):  # pylint: disable=unused-argument
        """
        A response was successful with this authentication. Reset retry count
        :return:
        """
        self.retry_count = 0

    def handle_401(self, headers):  # pylint: disable=unused-argument
        """
        :returns True if retry is allowed
        """
        self.retry_count += 1
        return self.retry_count <= self.max_retry

    def make_auth(self, method, url, headers):
        """
        Append authorization header
        """
        raise NotImplementedError


class BasicAuth(Auth):
    """
    Implementation of Basic authentication
    """

    def __init__(self, username, password, max_retry=1):
        super().__init__(max_retry)
        self.username = username
        self.password = password

    def make_auth(self, method, url, headers):
        b64 = b64encode(f'{self.username}:{self.password}'.encode())
        headers['Authorization'] = f'Basic {b64.decode()}'


class DigestAuth(Auth):
    """
    Implementation of Digest algorithm
    """

    def __init__(self, username, password, max_retry=1):
        super().__init__(max_retry)
        self.username = username
        self.password = password

        self.info = None

    @staticmethod
    def _parse_digest_header(header):
        fields = {}
        fields_ = parse_http_list(header)
        for field in fields_:
            k, v = field.split('=', 1)
            v = v.strip()
            if v and v[0] == v[-1] == '"':
                v = v[1:-1]
            fields[k.strip().lower()] = v
        return fields

    @staticmethod
    def _digest_function(algorithm: str) -> Callable[[str], str]:
        """
        Select the right digest function
        """
        assert algorithm in DIGEST_METHODS, f'algorithm {algorithm} not found'
        hashlib_digest = DIGEST_METHODS[algorithm]
        return lambda x: hashlib_digest(x.encode('utf-8')).hexdigest()

    def _prepare_digest_header(self, method: str, url: str) -> dict:
        """
        Prepare response header and return a dict; meant for ease of testing
        """

        assert self.info

        algorithm = self.info.get('algorithm', 'MD5').upper()
        realm = self.info.get('realm')
        nonce = self.info.get('nonce')
        opaque = self.info.get('opaque')

        hash_digest = self._digest_function(algorithm)

        A1 = '%s:%s:%s' % (self.username, realm, self.password)
        A2 = '%s:%s' % (method, url)

        HA1 = hash_digest(A1)
        HA2 = hash_digest(A2)

        # Direct response as per RFC 2069 - 2.1.1
        response = hash_digest(f'{HA1}:{nonce}:{HA2}')

        base = {
            'username': self.username,
            'realm': realm,
            'nonce': nonce,
            'uri': url,
            'response': response
        }

        if opaque:
            base['opaque'] = opaque

        return base

    def _build_digest_header(self, method: str, url: str) -> str:
        base = self._prepare_digest_header(method, url)

        opts = ", ".join(f'{k}="{v}"' for k, v in base.items())

        return f'Digest {opts}'

    def handle_401(self, headers: dict):
        """
        Takes the given response and tries digest-auth, if needed.

        :rtype: requests.Response
        """
        auth_header = headers['www-authenticate']
        if isinstance(auth_header, list):
            auth_header = next(header for header in auth_header if header.startswith('Digest '))
            # @TODO There may be several Digest propositions (MD5, SHA-256, ...)

        assert auth_header, 'unable to find a Digest header'
        self.info = self._parse_digest_header(auth_header[6:])

        return super().handle_401(headers)

    def handle_ok(self, headers: dict):
        """
        A response was successful with this authentication. Reset retry count
        :return:
        """
        if 'authentication-info' in headers:
            info = self._parse_digest_header(headers['authentication-info'])
            if 'nextnonce' in info:
                self.info['nonce'] = info['nextnonce']

        super().handle_ok(headers)

    def make_auth(self, method: str, url: str, headers: dict):
        """
        Add Authorization to the headers of given request
        :param method:
        :param url:
        :param headers:
        :return:
        """
        if self.info:
            headers['Authorization'] = self._build_digest_header(method, url)



"""
RTSP Media Session setup and control
"""
import asyncio
import calendar
import json
import logging
import math
import re
from datetime import datetime
from typing import Set
from urllib.parse import urlparse

from aiortsp.rtcp.stats import RTCPStats
from aiortsp.transport import RTPTransport
from .errors import RTSPError
from .parser import RTSPResponse
from .sdp import SDP

default_logger = logging.getLogger(__name__)


def sanitize_rtsp_url(url: str) -> str:
    """
    Sanitize an RTSP url, removing exotic scheme and authentication.
    """
    p_url = urlparse(url)
    scheme = p_url.scheme
    if scheme != 'rtsp' and scheme != 'rtsps':
        scheme = 'rtsp'
    return p_url._replace(
        scheme=scheme,
        netloc=f'{p_url.hostname}' if p_url.port is None else f'{p_url.hostname}:{p_url.port}'
    ).geturl()


class RTSPMediaSession:
    """
    RTSP Media Session
    TODO Refactor to support multiple medias
    """

    def __init__(self, connection, media_url, transport: RTPTransport, media_type='video', logger=None):
        self.connection = connection
        self.media_url = sanitize_rtsp_url(media_url)
        self.transport = transport
        self.media_type = media_type
        self.logger = logger or default_logger

        self.is_setup = False
        self.sdp = None
        self.server_rtp = None
        self.server_rtcp = None

        self.session_id = None
        self.session_keepalive = 60
        self.session_options: Set[str] = set()

    async def __aenter__(self):
        """
        At entrance of env, we expect the stream to be ready for playback
        """
        await self.setup()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_val and exc_type != asyncio.CancelledError:
            self.logger.error('exception during session: %s %s', exc_type, exc_val)
        await self.teardown()

    async def setup(self):
        """
        Perform SETUP
        """
        # Get supported options
        resp = await self._send('OPTIONS', url=self.media_url)
        self.save_options(resp)

        # Get SDP
        resp = await self._send('DESCRIBE', headers={
            'Accept': 'application/sdp'
        })

        if 'content-base' in resp.headers:
            self.media_url = resp.headers['content-base']
            self.logger.info('using base url: %s', self.media_url)

        self.logger.debug('received SDP:\n%s', resp.content)
        self.sdp = SDP(resp.content)
        self.logger.debug('parsed SDP:\n%s', json.dumps(self.sdp, indent=2))

        setup_url = self.sdp.setup_url(self.media_url, media_type=self.media_type)
        self.logger.info('setting up using URL: %s', setup_url)

        # --- SETUP <url> RTSP/1.0 ---
        headers = {}
        self.transport.on_transport_request(headers)
        resp = await self.connection.send_request('SETUP', url=setup_url, headers=headers)
        self.transport.on_transport_response(resp.headers)
        self.logger.info('stream correctly setup: %s', resp)

        # Store session ID
        self.save_session(resp)

        # Warm up transport
        await self.transport.warmup()

    @property
    def stats(self) -> RTCPStats:
        """Stats convenient accessor"""
        return self.transport.stats

    def save_options(self, resp: RTSPResponse):
        """
        Extract method lists from OPTIONS response
        """
        # Extract session Id
        if 'public' not in resp.headers:
            raise RTSPError('error on OPTIONS: `Public` not found')

        self.session_options = {o.strip().upper() for o in resp.headers['public'].split(',')}
        self.logger.info('session options: %s', self.session_options)

    def save_session(self, resp: RTSPResponse):
        """
        Extract session ID and timeout
        """
        # Extract session Id
        if 'session' not in resp.headers:
            raise RTSPError('error on SETUP: session not found')

        # Get session id
        session_params = resp.headers['session'].split(';')
        self.session_id = session_params[0].strip()
        timeout = 60
        if len(session_params) > 1:
            for option in session_params[1:]:
                option = option.strip()
                if not option.startswith('timeout'):
                    continue

                _, timeout_ = option.split('=', 1)
                timeout = int(timeout_)

        self.session_keepalive = int(timeout * 0.9)
        self.logger.info(
            'session id: %s, timeout: %s, keep_alive: %s',
            self.session_id, timeout, self.session_keepalive
        )

    async def teardown(self):
        """
        Perform TEARDOWN
        """
        if self.connection.running:
            self.logger.info('stopping session/playback...')
            resp = await self._send('TEARDOWN')
            self.logger.debug('response to teardown: %s', resp)
            return resp

        self.logger.info('session closed (no transport)')

    async def _send(self, method, url=None, headers=None):
        if headers is None:
            headers = {}
        if self.session_id:
            headers['Session'] = self.session_id
        return await self.connection.send_request(method, url or self.media_url, headers)

    @staticmethod
    def ts_to_clock(seek: float) -> str:
        """
        Must return a string in the following format:
            20190322T043720.003Z
        :param seek: utc timestamp
        """
        res = datetime.utcfromtimestamp(seek).strftime('%Y%m%dT%H%M%S')
        rem = seek - math.floor(seek)
        if rem:
            res += str(round(rem, 3))[1:]
        res += 'Z'
        return res

    @staticmethod
    def response_to_ts(resp, default_ts):
        """
        Try to return real play time, or default if not found
        """
        try:
            res = re.match(r'^ *clock *= *(?P<date>\d{8})T(?P<time>\d{6})(\.(?P<milli>\d+))?Z *-', resp.headers.get('range'))

            if not res:
                return default_ts

            content = res.groupdict()
            dt = datetime.strptime(content['date'] + content['time'], '%Y%m%d%H%M%S')

            ts = calendar.timegm(dt.timetuple())

            if content["milli"]:
                ts += float(f'0.{content["milli"]}')

            return ts
        except Exception:  # pylint: disable=broad-except
            # @TODO Should we log anything?
            return default_ts

    async def play(self, seek=None, speed=1):
        """
        Send a PLAY request
        :param seek: UTC timestamp where to ask to start. By default, uses 'now'.
        :param speed: Replay speed. Could be used for fast forward playing.
        """
        if seek:
            start = self.ts_to_clock(seek)
            range_ = f'clock={start}-'
        else:
            start = 'now'
            range_ = 'npt=now-'

        self.logger.info(
            'start playing %s at time `%s` and speed `%s`...',
            self.media_url, start, speed
        )

        resp = await self._send('PLAY', headers={
            'Scale': speed,
            'Range': range_
        })
        self.logger.debug('response to play: %s', resp)
        return resp

    async def pause(self):
        """
        Send a PAUSE, temporarily stopping RTP flow but keeping session alive.
        """
        self.logger.debug('sending keep alive')
        resp = await self._send('PAUSE')
        self.logger.debug('response to pause: %s', resp)
        return resp

    async def keep_alive(self):
        """
        Send a GET_PARAMETER or OPTIONS message in order to keep session alive.
        """
        self.logger.debug('sending keep alive')

        # We must send a supported command
        if 'GET_PARAMETER' in self.session_options:
            resp = await self._send('GET_PARAMETER')
        elif 'OPTIONS' in self.session_options:
            resp = await self._send('OPTIONS')
        else:
            self.logger.info('Does not support GET_PARAMETER or OPTIONS: not sending keep alive.')
            resp = None

        self.logger.debug('response to keep_alive: %s', resp)
        return resp



from binascii import hexlify

import pytest

from aiortsp.rtcp.parser import SR, SDES, RR, RTCP, ts_to_ntp, ntp_to_ts, BYE, SRReport
from aiortsp.rtcp.stats import RTCPStats


@pytest.mark.parametrize('data, klasses, alt', [
    ('80c8000677ae8d65e051bc2bea33b0001fa8034c0000000000000000', [SR], None),
    ('81c8000c5d931534dd3ac1b061d9c00000084f8000000d4c00084f8001932db40000000100000000000000000000000000000000', [SR], None),
    ('81ca000477ae8d650106756e6b6e6f7700000000', [SDES], None),
    ('81c9000730f5fb2730f5fb27000000000000726600000c0abc3ffbdb00030e42', [RR], None),
    ('81c9000730f5fb2730f5fb27000000000000726600000c0abc3ffbdb00030e4281ca000477ae8d650106756e6b6e6f7700000000', [RR, SDES], None),
    ('81cb0001166ae287', [BYE], None),

    # Padded
    ('a1ca000577ae8d650106756e6b6e6f770000000000000004', [SDES], '81ca000477ae8d650106756e6b6e6f7700000000'),
])
def test_rtcp_parse(data, klasses, alt):
    b = bytearray.fromhex(data)
    res = RTCP.unpack(b)

    assert [a.__class__ for a in res.packets] == klasses

    for k in klasses:
        assert isinstance(res.get(k.pt), k)

    # repack
    b2 = bytes(res)
    assert hexlify(b2).decode().lower() == (alt or data)

    # Just make sure the representation is not crashing
    print(res)


def test_ntp_to_time():
    t = 1554467594.1005
    n1, n2 = ts_to_ntp(t)
    assert t == ntp_to_ts(n1, n2)


def test_rtcp_stats():
    s = RTCPStats()
    # Mark first frame
    s.init_seq(65042)
    s.update_seq(65042)

    s.update_lost_expected()
    assert s.lost == 0
    assert s.maxseq == 65042
    assert s.extended_seq == 65042

    for i in range(10):
        s.update_seq(65043 + i)

    s.update_lost_expected()
    assert s.lost == 0
    assert s.maxseq == 65052
    assert s.extended_seq == 65052

    # Hole!
    s.update_seq(65060)

    s.update_lost_expected()
    assert s.lost == 7
    assert s.maxseq == 65060
    assert s.extended_seq == 65060

    # recover
    for i in range(7):
        s.update_seq(65053 + i)

    s.update_lost_expected()
    assert s.lost == 0
    assert s.maxseq == 65060
    assert s.extended_seq == 65060

    # Wrap!
    s.update_seq(2)

    s.update_lost_expected()
    assert s.lost == 2**16 - 65060 + 1
    assert s.maxseq == 2
    assert s.extended_seq == 2**16 + 2


def test_sr_report():

    sr = SRReport(
        ssrc=42, flost=0, clost=int(2**64),
        hseq=int(2**64), jitter=0,
        lsr=0, dlsr=0)

    assert len(bytes(sr)) > 0



import pytest

from aiortsp.rtsp.parser import RTSPParser


DATA = [
    # Very dumb one
    (b"""RTSP/1.0 200 OK\r\nCSeq: 0\r\n\r\n""", [('response', (200, 0, 0, None))]),

    # Response to play
    (b"""RTSP/1.0 200 OK\r
Range: clock=20190315T043720.003Z-\r
RTP-Info: url=rtsp://10.10.20.118:654/00000001-0000-babe-0000-accc8e000aa7/playback;ssrc=871954036;seq=60290;rtptime=0\r
CSeq: 3\r
Session: 35765835\r
X-RtspClientVersion: 5.7\r
\r\n""", [('response', (200, 3, 0, None))]),

    # SDP Answer
    (b"""RTSP/1.0 200 OK\r
Content-Type: application/sdp\r
Content-Length: 302\r
CSeq: 1\r
Session: 35765835\r
X-RtspClientVersion: 5.7\r
\r
v=0\r
o=- 0 0 IN IP4 0.0.0.0\r
s=\r
i=\r
c=IN IP4 0.0.0.0\r
t=0 0\r
m=video 0 RTP/AVP 96\r
a=rtpmap:96 H264/90000\r
a=fmtp:96 packetization-mode=1; profile-level-id=420029; sprop-parameter-sets=Z0IAKeKQFoJNgScFAQXh4kRU,aM48gA==\r
a=control:rtsp://10.10.20.118:654/00000001-0000-babe-0000-accc8e000aa7/playback\r
""", [('response', (200, 1, 302, """v=0\r
o=- 0 0 IN IP4 0.0.0.0\r
s=\r
i=\r
c=IN IP4 0.0.0.0\r
t=0 0\r
m=video 0 RTP/AVP 96\r
a=rtpmap:96 H264/90000\r
a=fmtp:96 packetization-mode=1; profile-level-id=420029; sprop-parameter-sets=Z0IAKeKQFoJNgScFAQXh4kRU,aM48gA==\r
a=control:rtsp://10.10.20.118:654/00000001-0000-babe-0000-accc8e000aa7/playback\r
"""))]),

    # Data not ending with an EOL
    (b"""RTSP/1.0 462 Destination Unreachable\r
Content-Type: text/plain\r
Content-Length: 110\r
CSeq: 2\r
Session: 64118488;timeout=90\r
X-RtspClientVersion: 5.7\r
\r
Client requested end point already in use for camera 00000001-0000-babe-0000-accc8e000aa7 in session 26378381.RTSP/1.0 200 OK\r\nCSeq: 3\r\n\r\n""", [
        ('response', (462, 2, 110, 'Client requested end point already in use for camera 00000001-0000-babe-0000-accc8e000aa7 in session 26378381.')),
        ('response', (200, 3, 0, None)),
    ]),

    # Multiple Answers
    (b"""RTSP/1.0 200 OK\r\nCSeq: 0\r\n\r\nRTSP/1.0 404 Not Found\r\nCSeq: 1\r\n\r\n""", [
        ('response', (200, 0, 0, None)),
        ('response', (404, 1, 0, None)),
    ]),

    # Multiple Answers, with extra lines in between! (not fully RFC compliant...)
    (b"""RTSP/1.0 200 OK\r\nCSeq: 0\r\n\r\n\r\nRTSP/1.0 404 Not Found\r\nCSeq: 1\r\n\r\n""", [
        ('response', (200, 0, 0, None)),
        ('response', (404, 1, 0, None)),
    ]),

    # Multiple binary
    (b"""$\000\000\003iii$\001\000\007Hello!!""", [
        ('binary', (0, b'iii')),
        ('binary', (1, b'Hello!!')),
    ]),

    # Binary followed by incomplete Response header
    (b"""$\000\000\003iiiRTSP/1.0 200 O""", [
        ('binary', (0, b'iii')),
    ]),

    # Answer and binary
    (b"""RTSP/1.0 200 OK\r\nCSeq: 0\r\n\r
$\000\000\003iii\r\n$\001\000\007Hello!!RTSP/1.0 404 Not Found\r\nCSeq: 1\r\n\r\n""", [
        ('response', (200, 0, 0, None)),
        ('binary', (0, b'iii')),
        ('binary', (1, b'Hello!!')),
        ('response', (404, 1, 0, None)),
    ]),

    # Announce incoming
    (b"""ANNOUNCE rtsp://foo.com/bar.avi RTSP/1.0\r
CSeq: 10\r
Session: 12345678\r
Require: method.announce\r
Event-Type: End-Of-Stream\r
Range: npt=10-100\r
RTP-Info: url= rtsp://foo.com/bar.avi/streamid=0; seq=456,\r
  url= rtsp://foo.com/bar.avi/streamid=1; seq=789\r
Content-Type: text/parameters\r
\r
""", [
        ('request', ('ANNOUNCE', 'rtsp://foo.com/bar.avi', 10)),
    ]),
]


@pytest.mark.parametrize('data, objs', DATA)
def test_simple_response(data, objs):

    parser = RTSPParser()

    resps = list(parser.parse(data))

    assert len(resps) == len(objs)

    for i in range(len(objs)):
        resp = resps[i]
        type_ = objs[i][0]

        assert resp.type == type_

        if type_ == 'response':
            status, cseq, length, content = objs[i][1]
            assert resp.status == status
            assert resp.cseq == cseq
            assert resp.content_length == length
            if content:
                assert resp.content == content
        elif type_ == 'binary':
            id_, data = objs[i][1]
            assert resp.id == id_
            assert resp.length == len(data)
            assert resp.data == data
        elif type_ == 'request':
            method, url, cseq = objs[i][1]
            assert resp.method == method
            assert resp.request_url == url
            assert resp.cseq == cseq


@pytest.mark.parametrize('data, objs', DATA)
def test_chunk_response(data, objs):

    parser = RTSPParser()

    resps = []

    for chunk_idx in range(1 + len(data) // 4):
        resps.extend(list(parser.parse(data[chunk_idx*4:(chunk_idx+1)*4])))

    assert len(resps) == len(objs)

    for i in range(len(objs)):
        resp = resps[i]
        type_ = objs[i][0]

        assert resp.type == type_

        if type_ == 'response':
            status, cseq, length, content = objs[i][1]
            assert resp.status == status
            assert resp.cseq == cseq
            assert resp.content_length == length
            if content:
                assert resp.content == content
        elif type_ == 'binary':
            id_, data = objs[i][1]
            assert resp.id == id_
            assert resp.length == len(data)
            assert resp.data == data
        elif type_ == 'request':
            method, url, cseq = objs[i][1]
            assert resp.method == method
            assert resp.request_url == url
            assert resp.cseq == cseq



"""
Simplified RTP reader
"""
import asyncio
import logging
import ssl
from time import time
from typing import AsyncIterable, Optional
from urllib.parse import urlparse

from dpkt.rtp import RTP

from aiortsp.rtsp.connection import RTSPConnection
from aiortsp.rtsp.session import RTSPMediaSession, sanitize_rtsp_url
from aiortsp.transport import transport_for_scheme, RTPTransport, RTPTransportClient

class RTSPReader(RTPTransportClient):
    """
    Quick wrapper around base functions to start getting frames from an RTSP feed.

    Usage:

    .. code-block::

        async with RTSPReader('rtsp://foo/bar') as reader:
            async for pkt in reader.iter_packets():
                print(pkt)
    """

    def __init__(
            self, media_url: str, timeout=10, log_level=20,
            ssl = None,
            run_loop=False, **_
    ):
        self.media_url = media_url
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(log_level)
        self.timeout = timeout
        self.run_loop = run_loop
        self.ssl = ssl
        self.queue: 'asyncio.Queue[RTP]' = asyncio.Queue()
        self._runner = None
        self.connection: Optional[RTSPConnection] = None
        self.transport: Optional[RTPTransport] = None
        self.session: Optional[RTSPMediaSession] = None
        self.payload_type = None

    def handle_rtp(self, rtp: RTP):
        """Queue packets for the iterator"""
        if self.payload_type and self.payload_type != rtp.pt:
            return

        self.queue.put_nowait(rtp)

    def on_ready(self, connection: RTSPConnection, transport: RTPTransport, session: RTSPMediaSession):
        """Handler on ready to play stream, for sub classes to do their initialisation"""
        if session.sdp:
            self.payload_type = session.sdp.media_payload_type()
        transport.subscribe(self)
        self.connection = connection
        self.transport = transport
        self.session = session

    def handle_closed(self, error):
        """Handler for connection closed, for sub classes to cleanup their state"""
        self.logger.info('connection closed, error: %s', error)
        self.connection = None
        self.transport = None
        self.session = None

    async def run_stream_loop(self):
        """Run stream as a loop, forever restarting unless if cancelled"""
        while True:
            try:
                await self.run_stream()
            except asyncio.CancelledError:
                self.logger.error('Stopping run loop for %s', sanitize_rtsp_url(self.media_url))
                break
            except Exception as ex:  # pylint: disable=broad-except
                self.logger.error('Error on stream: %r. Reconnecting...', ex)
                await asyncio.sleep(1)

    async def run_stream(self):
        """
        Setup and play stream, and ensure it stays on.
        """
        self.logger.info('try loading stream %s', sanitize_rtsp_url(self.media_url))

        p_url = urlparse(self.media_url)

        if p_url.scheme == 'rtsps' and not self.ssl:
            self.ssl = ssl.create_default_context()
        if p_url.scheme == 'rtsps':
            default_port = 322
        else:
            default_port = 554

        async with RTSPConnection(
                p_url.hostname, p_url.port or default_port,
                p_url.username, p_url.password,
                logger=self.logger, timeout=self.timeout,
                ssl=self.ssl
        ) as conn:
            self.logger.info('connected!')

            transport_class = transport_for_scheme(p_url.scheme)
            async with transport_class(conn, logger=self.logger, timeout=self.timeout) as transport:
                async with RTSPMediaSession(conn, self.media_url, transport=transport, logger=self.logger) as sess:

                    self.on_ready(conn, transport, sess)

                    self.logger.info('playing stream...')
                    await sess.play()

                    try:
                        last_keep_alive = time()
                        while conn.running and transport.running:
                            # Check keep alive
                            now = time()
                            if (now - last_keep_alive) > sess.session_keepalive:
                                await sess.keep_alive()
                                last_keep_alive = now

                            await asyncio.sleep(1)

                    except asyncio.CancelledError:
                        self.logger.info('stopping stream...')
                        raise

    async def __aenter__(self):
        self._runner = asyncio.ensure_future(
            self.run_stream_loop() if self.run_loop else self.run_stream())
        self._runner.add_done_callback(lambda *_: self.queue.put_nowait(None))
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._runner:
            self._runner.cancel()

    async def iter_packets(self) -> AsyncIterable[RTP]:
        """
        Yield RTP packets as they come.
        User can then do whatever they want, without too much boiler plate.
        """
        while True:
            pkt = await self.queue.get()

            if not pkt:
                break

            yield pkt



import pytest

from aiortsp.rtsp.parser import RTSPResponse, RTSPParser
from aiortsp.rtsp.session import RTSPMediaSession

CASES = [
    (1553609305.123, '20190326T140825.123Z'),
    (1553609305.12, '20190326T140825.12Z'),
    (1553609305.1, '20190326T140825.1Z'),
    (1553609305, '20190326T140825Z'),
]

@pytest.mark.parametrize('ref, exp', CASES)
def test_ts_to_clock(ref, exp):
    assert RTSPMediaSession.ts_to_clock(ref) == exp

@pytest.mark.parametrize('format', [
    'clock={}-',
    ' clock={}- ',
    ' clock ={} -',
    ' clock = {} -',
    ' clock =     {} -',
    ' clock = {} - 20190326T140828Z',
])
@pytest.mark.parametrize('exp, value', CASES)
def test_ts_to_clock(format, exp, value):
    resp = RTSPResponse()
    resp.headers = {
        'range': format.format(value)
    }
    assert RTSPMediaSession.response_to_ts(resp, None) == exp


def test_response_to_ts():

    parser = RTSPParser()

    resps = list(parser.parse(b"""RTSP/1.0 200 OK\r
CSeq: 4\r
Session: FD8A666F\r
Range: clock=20180101T010203.045Z-\r
RTP-Info: url=rtsp://192.168.1.98/axis-media/media.amp/trackID=1;seq=44350;rtptime=31576935\r
Date: Wed, 30 Sep 2020 20:08:49 GMT\r
\r
"""))

    assert len(resps) == 1
    assert RTSPMediaSession.response_to_ts(resps[0], 1) == 1514768523.045


import setuptools
import pathlib

root_dir = pathlib.Path(__file__).parent
with (root_dir / 'aiortsp' / '__version__.py').open(encoding='utf-8') as f:
    exec(f.read())

setuptools.setup(
    version=__version__,
)



"""
RTSP Stream parsing module
"""
from abc import abstractmethod
from binascii import hexlify
from io import BytesIO
from typing import Tuple, Iterator

from .errors import RTSPResponseError

CRLF = b'\r\n'


class RTSPMessage:
    """
    Base return class for any parsed object coming from an RTSP connection.
    """

    type = 'unknown'

    @abstractmethod
    def feed(self, data: bytes) -> Tuple[bytes, bool]:
        """
        Feed part with a line.
        :returns remaining bytes and True if done parsing current packet
        """

    @property
    def content(self) -> str:
        """Extract data from buffer as utf-8"""
        raise NotImplementedError

    @property
    def data(self) -> bytes:
        """Extract data from buffer as bytes"""
        raise NotImplementedError

    @property
    def length(self) -> int:
        """Return content length"""
        raise NotImplementedError


class HTTPLikeMsg(RTSPMessage):
    """
    Base class for both RTSP request & reply. Only the first line differs.
    """

    def __init__(self, buffer_size=2**16):
        self.headerlist = []
        self.headers = None
        self._untreated_data = b''
        self._data = None
        self.size = 0
        self._buf = ''
        self.buffer_size = buffer_size
        self.cseq = None
        self.content_type = None
        self.content_length = -1

        self.first_line = None

    @abstractmethod
    def parse_first_line(self, line: str):
        """
        Parse the first line received
        """

    def feed(self, data: bytes) -> Tuple[bytes, bool]:
        """
        Feed part with a line.
        :returns remaining bytes and True if done parsing current packet
        """
        done = False

        # Prepend any leftover from previous message
        data, self._untreated_data = self._untreated_data + data, b''

        while data and not done:
            if not self._data:
                # Still a header
                if b'\r\n' not in data:
                    # We don't have yet a full header: will have next time (hopefully)
                    self._untreated_data = data
                    data = b''
                    break
                line, data = data.split(b'\r\n', 1)
                done = self.parse_header(line.decode('utf-8') + '\r\n')
            else:
                data, done = self.parse_body(data)

        return data, done

    def parse_header(self, line) -> bool:
        """
        Add a header line
        :returns True if message is finished (no content)
        """
        if self.first_line is None:
            self.parse_first_line(line)
            return False

        line = line.strip()
        if not line:  # blank line -> end of header segment
            return self.finish_header()

        if line[0] in ' \t' and self.headerlist:
            name, value = self.headerlist.pop()
            self.headerlist.append((name, value + line.strip()))
        else:
            if ':' not in line:
                raise RTSPResponseError("Syntax error in header: No colon.")
            name, value = line.split(':', 1)
            self.headerlist.append((name.strip().lower(), value.strip()))
        return False

    def parse_body(self, data: bytes) -> Tuple[bytes, bool]:
        """
        Add data to the body.
        :returns (data, done) with:
            - data as leftover bytes after end of body
            - done a boolean True when body is complete
        """
        # Do we have too much data?
        rem_data = self.content_length - self.size
        assert rem_data > 0, 'we should not be here if already done'

        if len(data) > rem_data:
            leftover = data[rem_data:]
            data = data[:rem_data]
        else:
            leftover = b''

        self.size += len(data)
        self._data.write(data)

        if self.size > self.buffer_size:
            raise RTSPResponseError('Size of body exceeds maximum buffer size')

        return leftover, self.size == self.content_length

    def finish_header(self) -> bool:
        """Last header received; create buffer and extra useful info"""
        self.headers = {}
        for k, v in self.headerlist:
            if k in self.headers:
                if not isinstance(self.headers[k], list):
                    self.headers[k] = [self.headers[k]]
                self.headers[k].append(v)
            else:
                self.headers[k] = v
        self.content_type = self.headers.get('content-type', self.content_type)
        self.content_length = int(self.headers.get('content-length', '0'))
        self.cseq = int(self.headers.get('cseq', '-1'))

        has_content = self.content_length > 0

        if has_content:  # soon: if has_content := self.content_length > 0
            self._data = BytesIO()

        return not has_content

    @property
    def length(self) -> int:
        """Data length"""
        return len(self.data)

    @property
    def data(self) -> bytes:
        """Extract data from buffer as bytes"""
        val = b''
        if self._data:
            pos = self._data.tell()
            self._data.seek(0)
            try:
                val = self._data.read()
            finally:
                self._data.seek(pos)
        return val

    @property
    def content(self) -> str:
        """Extract data from buffer as utf-8"""
        return self.data.decode('utf-8')


class RTSPRequest(HTTPLikeMsg):
    """
    RTSP Request parser.
    """

    type = 'request'
    CLIENT_REQUESTS = (
        b'OPTIONS',
        b'DESCRIBE',
        b'ANNOUNCE',
        b'SETUP',
        b'PLAY',
        b'PAUSE',
        b'TEARDOWN',
        b'GET_PARAMETER',
        b'SET_PARAMETER',
        b'REDIRECT',
        b'RECORD',
    )

    def __init__(self, buffer_size=2 ** 16):
        super().__init__(buffer_size)

        self.request_url = None
        self.method = None

    def parse_first_line(self, line):
        self.first_line = line
        self.method, self.request_url, protocol = line.split(None, 2)
        assert protocol.strip().startswith('RTSP/1.0'), 'RTSP response should start with an RTSP/1.0 protocol marker'

    def __repr__(self):
        return f'<Request ' \
            f'type={self.method} ' \
            f'url={self.request_url} ' \
            f'headers={self.headers} ' \
            f'content-length={self.content_length}>'


class RTSPResponse(HTTPLikeMsg):
    """
    RTSP Response parser.
    """

    type = 'response'

    def __init__(self, buffer_size=2 ** 16):
        super().__init__(buffer_size)

        self.status = None
        self.status_msg = None

    def parse_first_line(self, line):
        assert line.startswith('RTSP/1.0'), 'RTSP response should start with an RTSP/1.0 protocol marker'
        self.first_line = line
        _, status, status_msg = line.split(None, 2)
        self.status = int(status.strip())
        self.status_msg = status_msg.strip()

    def __repr__(self):
        return f'<Response ' \
            f'status={self.status} ' \
            f'msg="{self.status_msg}"" ' \
            f'headers={self.headers} ' \
            f'content-length={self.content_length}>'


class RTSPBinary:
    """
    RTSP inline Binary parser.
    """

    type = 'binary'

    def __init__(self):
        self._buf = b''

    def feed(self, data):
        """
        Feed part with a line.
        :returns a frame if done
        """
        self._buf += data

        # If we did not get enough to read id and length, just return
        if len(self._buf) < 4:
            return b'', False

        if len(self._buf) < (self.length + 4):
            return b'', False

        data = self._buf[self.length + 4:]
        self._buf = self._buf[:self.length + 4]
        return data, True

    @property
    def id(self) -> int:
        """
        RTSP inline binary data all have a reference ID,
        negociated during SETUP.
        """
        assert len(self._buf) >= 1, 'buffer too short'
        return self._buf[1]

    @property
    def length(self) -> int:
        """Return frame length"""
        assert len(self._buf) >= 4, 'buffer too short'
        return (self._buf[2] << 8) + self._buf[3]

    @property
    def content(self) -> str:
        """Return string representation of content"""
        return hexlify(self.data).decode('utf-8')

    @property
    def data(self) -> bytes:
        """Extract data from buffer"""
        return self._buf[4:]

    def __repr__(self):
        return f'<Binary id={self.id} length={self.length}>'


class RTSPParser:
    """
    RTSP Stream parser.
    Keep track if incoming data and yields valid elements upon calls to parse().
    """

    def __init__(self):
        self.pending_msg = None
        self._prev_data = b''

    def parse(self, data: bytes) -> Iterator[RTSPMessage]:
        """
        Receiving data
        :param data:
        :return:
        """
        while data:
            if self.pending_msg is None:
                # Prepend potential leftover from previous data
                data, self._prev_data = self._prev_data + data, b''

                # We need to determine what is coming next
                if data.startswith(CRLF):
                    # Skip empty lines between items
                    data = data[2:]
                    continue

                if data.startswith(b'RTSP'):
                    # This is a reply
                    self.pending_msg = RTSPResponse()

                elif data.startswith(b'$'):
                    self.pending_msg = RTSPBinary()

                elif data.startswith(RTSPRequest.CLIENT_REQUESTS) or any(
                        req.startswith(data) for req in RTSPRequest.CLIENT_REQUESTS):
                    self.pending_msg = RTSPRequest()

                elif len(data) > 13:
                    # That's the longest client request we could expect...
                    raise ValueError

                else:
                    # Received only a chunk of message. Should be better next iteration. Store
                    self._prev_data = data
                    break

            data, done = self.pending_msg.feed(data)

            if done:
                yield self.pending_msg
                self.pending_msg = None



import asyncio
import contextlib

import logging

from aiortsp.transport import RTPTransportClient, transport_for_scheme
from aiortsp.rtsp.connection import RTSPConnection
from aiortsp.rtsp.session import RTSPMediaSession

logger = logging.getLogger('rtsp_client')
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s')


class Probe(RTPTransportClient):

    def __init__(self):
        self.rtp_count = 0
        self.rtcp_count = 0

    def handle_rtp(self, rtp):
        self.rtp_count += 1

    def handle_rtcp(self, rtcp):
        self.rtcp_count += 1
        logger.debug('RTCP received: %s', rtcp)


async def main():
    import argparse
    from urllib.parse import urlparse

    parser = argparse.ArgumentParser()
    parser.add_argument('-l', '--logging', type=int, default=20, help='RTSP url')
    parser.add_argument('-a', '--address', default='127.0.0.1', help='destination UDP address')
    parser.add_argument('-A', '--auth', type=str, help='Auth to force ')
    parser.add_argument('-p', '--props', default=None, help='Stream props (guessed if not provided)')
    parser.add_argument('-t', '--timeout', type=int, default=10, help='UDP timeout')
    parser.add_argument('url', help='RTSP url')
    args = parser.parse_args()

    logger.setLevel(args.logging)

    p_url = urlparse(args.url)
    media_url = args.url
    probe = Probe()

    async with RTSPConnection(
            p_url.hostname,
            p_url.port or 554,
            p_url.username,
            p_url.password,
            logger=logger
    ) as conn:

        logger.info('connected!')

        # Detects if UDP or TCP must be used for RTP transport
        transport_class = transport_for_scheme(p_url.scheme)

        async with transport_class(conn, logger=logger, timeout=args.timeout) as transport:

            # This is where wa actually subscribe to data
            transport.subscribe(probe)

            async with RTSPMediaSession(conn, media_url, transport, logger=logger) as sess:

                await sess.play()

                try:
                    while conn.running and transport.running:
                        await asyncio.sleep(sess.session_keepalive)
                        await sess.keep_alive()

                        logger.info('received %s RTP, %s RTCP', probe.rtp_count, probe.rtcp_count)

                except asyncio.CancelledError:
                    logger.info('stopping stream...')


if __name__ == '__main__':
    with contextlib.suppress(KeyboardInterrupt):
        asyncio.run(main())



import pytest

from aiortsp.rtsp.sdp import SDP


def test_sdp_simple():

    sdp = SDP("""v=0\r
o=- 0 0 IN IP4 0.0.0.0\r
s=\r
i=\r
c=IN IP4 0.0.0.0\r
t=0 0\r
m=video 0 RTP/AVP             96\r
a=rtpmap:96 H264/90000\r
a=fmtp:96  packetization-mode=1 ; profile-level-id=420029 ; sprop-parameter-sets=Z0IAKeKQFoJNgScFAQXh4kRU,aM48gA==  \r
a=control:rtsp://10.10.20.118:654/00000001-0000-babe-0000-accc8e000aa7/playback\r
""")

    assert sdp['version'] == 0
    assert len(sdp['medias']) == 1
    med = sdp['medias'][0]

    assert med == sdp.get_media()

    assert med['protocol'] == 'RTP/AVP'

    attrs = med['attributes']

    # Control
    assert 'control' in attrs
    assert attrs['control'] == 'rtsp://10.10.20.118:654/00000001-0000-babe-0000-accc8e000aa7/playback'

    # fmtp
    assert 'fmtp' in attrs
    assert attrs['fmtp']['pt'] == 96
    assert attrs['fmtp']['sprop-parameter-sets'] == 'Z0IAKeKQFoJNgScFAQXh4kRU,aM48gA=='
    assert attrs['fmtp']['profile-level-id'] == '420029'
    assert attrs['fmtp']['packetization-mode'] == '1'


def test_sdp_errors():
    """
    Add specially crafter errors which we want the parser to be robust about:
        - a trailing `;` at the end of `fmtp`
        - an empty line at the end and garbage
    """

    sdp = SDP("""v=0\r
o=- 0 0 IN IP4 0.0.0.0\r
s=\r
i=\r
c=IN IP4 0.0.0.0\r
t=0 0\r
m=video 0 RTP/AVP             96\r
a=rtpmap:96 H264/90000\r
a=fmtp:96  packetization-mode=1 ; profile-level-id=420029 ; sprop-parameter-sets=Z0IAKeKQFoJNgScFAQXh4kRU,aM48gA== ; \r
a=control:rtsp://10.10.20.118:654/00000001-0000-babe-0000-accc8e000aa7/playback\r
hello=world\r
\r
""")

    assert sdp['version'] == 0
    assert len(sdp['medias']) == 1
    med = sdp['medias'][0]

    assert med == sdp.get_media()

    assert med['protocol'] == 'RTP/AVP'

    attrs = med['attributes']

    # Control
    assert 'control' in attrs
    assert attrs['control'] == 'rtsp://10.10.20.118:654/00000001-0000-babe-0000-accc8e000aa7/playback'

    # fmtp
    assert 'fmtp' in attrs
    assert attrs['fmtp']['pt'] == 96
    assert attrs['fmtp']['sprop-parameter-sets'] == 'Z0IAKeKQFoJNgScFAQXh4kRU,aM48gA=='
    assert attrs['fmtp']['profile-level-id'] == '420029'
    assert attrs['fmtp']['packetization-mode'] == '1'


@pytest.mark.parametrize('mix_ctrl, video_ctrl, req_base, expected_url', [
    # When no mix URL is provided but relative stream control is, mix base and stream
    ('*', 'stream=0', 'rtsp://foo/bar/', 'rtsp://foo/bar/stream=0'),
    # With a mix URL and relative stream control, base URL should be overwritten and aggregated
    ('rtsp://foo/toto/', 'stream=0', 'rtsp://foo/bar/', 'rtsp://foo/toto/stream=0'),

    # Same with absolute URLs
    ('*', 'rtsp://foo/snake/stream=0', 'rtsp://foo/bar/', 'rtsp://foo/snake/stream=0'),
    ('rtsp://foo/toto/', 'rtsp://foo/snake/stream=0', 'rtsp://foo/bar/', 'rtsp://foo/snake/stream=0'),

    # In some cases, a trailing backslash is not provided while a relative stream is...
    ('rtsp://foo/toto', 'stream=0', 'rtsp://foo/bar/', 'rtsp://foo/toto/stream=0'),
    ('*', 'stream=0', 'rtsp://foo/bar', 'rtsp://foo/bar/stream=0'),
])
def test_sdp_control_generic(mix_ctrl, video_ctrl, req_base, expected_url):
    sdp = SDP(f"""v=0\r
o=- 0 0 IN IP4 0.0.0.0\r
s=\r
i=\r
c=IN IP4 0.0.0.0\r
t=0 0\r
a=control:{mix_ctrl}\r
m=video 0 RTP/AVP             96\r
a=rtpmap:96 H264/90000\r
a=fmtp:96  packetization-mode=1 ; profile-level-id=420029 ; sprop-parameter-sets=Z0IAKeKQFoJNgScFAQXh4kRU,aM48gA==  \r
a=control:{video_ctrl}\r
""")
    assert sdp.setup_url(req_base) == expected_url


